from typing import List, Optional
from pydantic import BaseModel
from pathlib import Path
from abc import ABC, abstractmethod
import json


class ForeignKey(BaseModel):
    table: str
    column: str


class Column(BaseModel):
    name: str
    type: str
    is_primary_key: bool = False
    foreign_key: Optional[ForeignKey] = None

    def __hash__(self):
        return hash((self.name, self.type, self.is_primary_key, self.foreign_key))


class Table(BaseModel):
    name: str
    columns: List[Column]
    data: List[dict] = []


class ITablesSnapshot(ABC):
    @abstractmethod
    def get_table(self, name: str) -> Optional[Table]:
        raise NotImplementedError("get_table method must be implemented")


class InMemoryTables(BaseModel, ITablesSnapshot):
    tables: List[Table]

    def get_table(self, name: str) -> Optional[Table]:
        for table in self.tables:
            if table.name == name:
                return table
        return None


class FileSystemTables(ITablesSnapshot):
    workdir: Path

    def __init__(self, workdir: Path):
        self.workdir = workdir

    def get_table(self, name: str):
        table_path = self.workdir / f"{name}.json"
        if not table_path.exists():
            raise FileNotFoundError(f"File {table_path} does not exist")
        rows = json.loads(table_path.read_text())
        columns = set()
        for row in rows:
            assert isinstance(row, dict), f"Row {row} is not a dictionary"
            for key, value in row.items():
                if key not in columns:
                    col = Column(name=key, type=type(value).__name__)
                    columns.add(col)
        return Table(name=name, columns=list(columns), data=rows)



from typing import List, Dict, Optional
from .tables import ITablesSnapshot
from .field_name import field_name
from .ast import (
    Expression,
    StringExpression,
    IntExpression,
    Select,
    FloatExpression,
    NameExpression,
    From,
    SelectField,
    Where,
    GroupBy,
    Command,
    PlusExpression,
    OrderBy,
    MinusExpression,
    MultiplyExpression,
    DivideExpression,
    EqualExpression,
    NotEqualExpression,
    GreaterThanExpression,
    GreaterThanOrEqualExpression,
    LessThanExpression,
    LessThanOrEqualExpression,
    Limit,
)


def apply_expression(expression: Expression, ctx: dict):
    if isinstance(expression, StringExpression):
        return expression.value
    elif isinstance(expression, IntExpression):
        return expression.value
    elif isinstance(expression, FloatExpression):
        return expression.value
    elif isinstance(expression, NameExpression) and expression.name.lower() in [
        "true",
        "false",
    ]:
        return expression.name.lower() == "true"
    elif isinstance(expression, NameExpression):
        if expression.name in ctx:
            return ctx[expression.name]
        else:
            raise ValueError(f"Unknown variable: {expression.name}")
    elif isinstance(expression, PlusExpression):
        left_value = apply_expression(expression.left, ctx)
        right_value = apply_expression(expression.right, ctx)
        if isinstance(left_value, (int, float)) and isinstance(
            right_value, (int, float)
        ):
            return left_value + right_value
        else:
            raise ValueError(
                f"Unsupported types for addition: {type(left_value)}, {type(right_value)}"
            )
    elif isinstance(expression, MinusExpression):
        left_value = apply_expression(expression.left, ctx)
        right_value = apply_expression(expression.right, ctx)
        if isinstance(left_value, (int, float)) and isinstance(
            right_value, (int, float)
        ):
            return left_value - right_value
        else:
            raise ValueError(
                f"Unsupported types for subtraction: {type(left_value)}, {type(right_value)}"
            )
    elif isinstance(expression, MultiplyExpression):
        left_value = apply_expression(expression.left, ctx)
        right_value = apply_expression(expression.right, ctx)
        if isinstance(left_value, (int, float)) and isinstance(
            right_value, (int, float)
        ):
            return left_value * right_value
        else:
            raise ValueError(
                f"Unsupported types for multiplication: {type(left_value)}, {type(right_value)}"
            )
    elif isinstance(expression, DivideExpression):
        left_value = apply_expression(expression.left, ctx)
        right_value = apply_expression(expression.right, ctx)
        if isinstance(left_value, (int, float)) and isinstance(
            right_value, (int, float)
        ):
            if right_value == 0:
                raise ValueError("Division by zero")
            return left_value / right_value
        else:
            raise ValueError(
                f"Unsupported types for division: {type(left_value)}, {type(right_value)}"
            )
    elif isinstance(expression, EqualExpression):
        left_value = apply_expression(expression.left, ctx)
        right_value = apply_expression(expression.right, ctx)
        if left_value == right_value:
            return True
        else:
            return False
    elif isinstance(expression, NotEqualExpression):
        left_value = apply_expression(expression.left, ctx)
        right_value = apply_expression(expression.right, ctx)
        if left_value != right_value:
            return True
        else:
            return False
    elif isinstance(expression, GreaterThanExpression):
        left_value = apply_expression(expression.left, ctx)
        right_value = apply_expression(expression.right, ctx)
        if isinstance(left_value, (int, float)) and isinstance(
            right_value, (int, float)
        ):
            return left_value > right_value
        else:
            raise ValueError(
                f"Unsupported types for greater than: {type(left_value)}, {type(right_value)}"
            )
    elif isinstance(expression, GreaterThanOrEqualExpression):
        left_value = apply_expression(expression.left, ctx)
        right_value = apply_expression(expression.right, ctx)
        if isinstance(left_value, (int, float)) and isinstance(
            right_value, (int, float)
        ):
            return left_value >= right_value
        else:
            raise ValueError(
                f"Unsupported types for greater than or equal: {type(left_value)}, {type(right_value)}"
            )
    elif isinstance(expression, LessThanExpression):
        left_value = apply_expression(expression.left, ctx)
        right_value = apply_expression(expression.right, ctx)
        if isinstance(left_value, (int, float)) and isinstance(
            right_value, (int, float)
        ):
            return left_value < right_value
        else:
            raise ValueError(
                f"Unsupported types for less than: {type(left_value)}, {type(right_value)}"
            )
    elif isinstance(expression, LessThanOrEqualExpression):
        left_value = apply_expression(expression.left, ctx)
        right_value = apply_expression(expression.right, ctx)
        if isinstance(left_value, (int, float)) and isinstance(
            right_value, (int, float)
        ):
            return left_value <= right_value
        else:
            raise ValueError(
                f"Unsupported types for less than or equal: {type(left_value)}, {type(right_value)}"
            )
    else:
        raise ValueError(f"Unsupported expression type: {type(expression)}")


def apply_where(where: Where, data: List[dict], ctx: dict):
    result = []
    for row in data:
        value = apply_expression(where.expression, {**ctx, **row})
        if value is True:
            result.append(row)
        elif value is False:
            continue
        else:
            raise ValueError(f"Where expressions should return bool, not {value}")
    return result


def apply_order_by(order_by: OrderBy, data: List[dict], ctx: dict):
    for order_field in order_by.fields:
        data.sort(
            key=lambda x: apply_expression(order_field.expression, {**ctx, **x}),
            reverse=(order_field.direction == "DESC"),
        )
    return data


def apply_group_by(group_by: GroupBy, data: List[dict], ctx: dict):
    groups: Dict[tuple, list] = {}
    for row in data:
        key = tuple(
            apply_expression(field, {**ctx, **row}) for field in group_by.fields
        )
        if key not in groups:
            groups[key] = []
        groups[key].append(row)
    return groups


def apply_limit(limit: Limit, data: List[dict], ctx: dict):
    start = limit.offset
    end = start + limit.limit
    return data[start:end]


def apply_select_fields(fields: List[SelectField], data: List[dict], ctx: dict):
    return [
        {
            field_name(field) or field.expression: apply_expression(
                field.expression, {**ctx, **row}
            )
            for field in fields
        }
        for row in data
    ]


def apply_from(
    from_part: Optional[From], tables: ITablesSnapshot, ctx: dict
) -> List[dict]:
    if from_part is None:
        return [{}]
    table = tables.get_table(from_part.table)
    if not table:
        raise ValueError(f"Table {from_part.table} not found")
    data = table.data
    if from_part.join:
        for join in from_part.join:
            join_table = tables.get_table(join.table)
            if not join_table:
                raise ValueError(f"Table {join.table} not found")
            data = [
                {**row, **join_row}
                for row in data
                for join_row in join_table.data
                if apply_expression(join.on, {**ctx, **row, **join_row})
            ]
    return data


def apply_select(select: Select, tables: ITablesSnapshot, ctx: dict):
    data = apply_from(select.from_part, tables, ctx)
    if select.where_part:
        data = apply_where(select.where_part, data, ctx)
    if select.group_part:
        data = apply_group_by(select.group_part, data, ctx)
    if select.order_part:
        data = apply_order_by(select.order_part, data, ctx)
    if select.limit_part:
        data = apply_limit(select.limit_part, data, ctx)
    if select.field_parts:
        data = apply_select_fields(select.field_parts, data, ctx)
    return data


def apply_command(command: Command, tables: ITablesSnapshot, ctx: dict):
    if isinstance(command, Select):
        return apply_select(command, tables, ctx)
    else:
        raise ValueError(f"Unsupported command type: {type(command)}")



from pathlib import Path
from .tables import FileSystemTables
from .eval import eval_sql
from argparse import ArgumentParser
from csv import DictWriter
from sys import stdout
from typing import List, Literal
from json import dumps


def query(code: str, workdir: Path, ctx: dict):
    tables = FileSystemTables(workdir=workdir)
    result = eval_sql(code=code, tables=tables, ctx=ctx)
    return result


def print_result(result: List[dict], format: Literal["json", "csv"]):
    if format == "json":
        print(dumps(result, indent=2))
    elif format == "csv":
        if len(result) == 0:
            print("No results to write to CSV.")
        else:
            keys = result[0].keys()
            writer = DictWriter(stdout, fieldnames=keys)
            writer.writeheader()
            for row in result:
                writer.writerow(row)


def main():
    parser = ArgumentParser(description="Run SQL queries on JSON files.")
    parser.add_argument("--code", type=str, help="SQL query to execute", default=None)
    parser.add_argument(
        "--workdir",
        type=Path,
        default=Path.cwd(),
        help="Directory containing JSON files",
    )
    parser.add_argument(
        "--ctx",
        type=str,
        default="{}",
        help="Context for the query (default: empty dictionary)",
    )
    parser.add_argument(
        "--format",
        type=str,
        default="json",
        choices=["json", "csv"],
        help="Output format (default: json)",
    )
    args = parser.parse_args()

    if args.code:
        result = query(code=args.code, workdir=args.workdir, ctx={})
        print_result(result=result, format=args.format)
    else:
        print("JSON SQL CLI")
        print("Type 'exit' to quit.")
        while True:
            try:
                code = input("> ")
                if code.lower() == "exit":
                    break
                result = query(code=code, workdir=args.workdir, ctx={})
                print_result(result=result, format=args.format)
            except KeyboardInterrupt:
                break
            except EOFError:
                break
            except SystemExit:
                break
            except Exception as e:
                print(f"Error: {e}")
                continue


if __name__ == "__main__":
    main()



from .tokens import Token
from typing import List, Tuple
from .ast import (
    Ast,
    Select,
    SelectField,
    Limit,
    IntExpression,
    SelectWildcard,
    From,
    NameExpression,
    EqualExpression,
    FloatExpression,
    LessThanOrEqualExpression,
    NotEqualExpression,
    GreaterThanExpression,
    LessThanExpression,
    GreaterThanOrEqualExpression,
    Where,
    StringExpression,
    OrderBy,
    OrderField,
    PlusExpression,
    MinusExpression,
    MultiplyExpression,
    DivideExpression,
    Expression,
)


def parse_order(tokens: List[Token]) -> Tuple[OrderBy, List[Token]]:
    if not tokens:
        return None, tokens

    assert tokens[0].type == "keyword" and tokens[0].value.upper() == "ORDER BY", (
        "Expected ORDER BY statement, got: " + str(tokens[0])
    )
    tokens = tokens[1:]
    if not tokens:
        return None, tokens

    order_fields: List[OrderField] = []
    while tokens and tokens[0].type != "comma":
        exp, tokens = parse_expression(tokens)
        if (
            tokens
            and tokens[0].type == "keyword"
            and tokens[0].value.upper()
            in [
                "ASC",
                "DESC",
            ]
        ):
            direction = tokens[0].value.upper()
            tokens = tokens[1:]
        else:
            direction = "ASC"
        order_fields.append(OrderField(expression=exp, direction=direction))

    return OrderBy(fields=order_fields), tokens


def parse_expression(tokens: List[Token]) -> Tuple[Expression, List[Token]]:
    stack = []
    while tokens:
        next_token = tokens[0]
        tokens = tokens[1:]
        if next_token.type == "int":
            stack.append(IntExpression(value=int(next_token.value)))
        elif next_token.type == "float":
            stack.append(FloatExpression(value=float(next_token.value)))
        elif next_token.type == "str":
            stack.append(StringExpression(value=next_token.value))
        elif next_token.type == "name":
            stack.append(NameExpression(name=next_token.value))
        elif next_token.type == "operator":
            operator = next_token.value
            if operator == "+":
                right, tokens = parse_expression(tokens)
                left = stack.pop()
                stack.append(PlusExpression(left=left, right=right))
            elif operator == "-":
                right, tokens = parse_expression(tokens)
                left = stack.pop()
                stack.append(MinusExpression(left=left, right=right))
            elif operator == "*":
                right, tokens = parse_expression(tokens)
                left = stack.pop()
                stack.append(MultiplyExpression(left=left, right=right))
            elif operator == "/":
                right, tokens = parse_expression(tokens)
                left = stack.pop()
                stack.append(DivideExpression(left=left, right=right))
            elif operator == "=":
                right, tokens = parse_expression(tokens)
                left = stack.pop()
                stack.append(EqualExpression(left=left, right=right))
            elif operator == "<":
                right, tokens = parse_expression(tokens)
                left = stack.pop()
                stack.append(LessThanExpression(left=left, right=right))
            elif operator == "<=":
                right, tokens = parse_expression(tokens)
                left = stack.pop()
                stack.append(LessThanOrEqualExpression(left=left, right=right))
            elif operator == ">":
                right, tokens = parse_expression(tokens)
                left = stack.pop()
                stack.append(GreaterThanExpression(left=left, right=right))
            elif operator == ">=":
                right, tokens = parse_expression(tokens)
                left = stack.pop()
                stack.append(GreaterThanOrEqualExpression(left=left, right=right))
            elif operator == "!=" or operator == "<>":
                right, tokens = parse_expression(tokens)
                left = stack.pop()
                stack.append(NotEqualExpression(left=left, right=right))
            else:
                raise ValueError(f"Unknown operator: {operator}")
        else:
            tokens = [next_token] + tokens
            break
    if len(stack) == 1:
        return stack[0], tokens
    else:
        raise ValueError("Invalid expression: " + str(stack))


def parse_where(tokens: List[Token]) -> Tuple[Where, List[Token]]:
    if not tokens or tokens[0].type != "keyword" or tokens[0].value.upper() != "WHERE":
        return None, tokens
    tokens = tokens[1:]
    expression, tokens = parse_expression(tokens)
    return Where(expression=expression), tokens


def parse_from(tokens: List[Token]) -> Tuple[From, List[Token]]:
    if len(tokens) == 0:
        return None, tokens
    if tokens[0].type != "keyword" or tokens[0].value.upper() != "FROM":
        raise ValueError("Expected FROM statement")

    tokens = tokens[1:]
    table = tokens[0]
    assert table.type == "name", f"Expected table name, got {table}"

    tokens = tokens[1:]

    if (
        len(tokens) > 0
        and tokens[0].type == "keyword"
        and tokens[0].value.upper() == "AS"
    ):
        tokens = tokens[1:]
        alias_token = tokens[0]
        assert alias_token.type == "name", f"Expected alias name, got {alias_token}"
        tokens = tokens[1:]
        return From(table=table.value, alias=alias_token.value), tokens
    else:
        return From(table=table.value), tokens


def parse_fields(tokens: List[Token]) -> Tuple[List[SelectField], List[Token]]:
    fields: List[SelectField] = []
    while tokens and tokens[0].type != "comma":
        if tokens[0].type == "keyword" and tokens[0].value.upper() == "FROM":
            break
        if tokens[0].type == "wildcard":
            fields.append(SelectWildcard())
            tokens = tokens[1:]
        else:
            exp, tokens = parse_expression(tokens)
            field = SelectField(expression=exp)
            if (
                tokens
                and tokens[0].type == "keyword"
                and tokens[0].value.upper() == "AS"
            ):
                tokens = tokens[1:]
                alias_token = tokens[0]
                assert (
                    alias_token.type == "name"
                ), f"Expected alias name, got {alias_token}"
                field.alias = alias_token.value
                tokens = tokens[1:]
            fields.append(field)
    return fields, tokens


def parse_limit(tokens: List[Token]) -> Tuple[Limit, List[Token]]:
    if not tokens or tokens[0].type != "keyword" or tokens[0].value.upper() != "LIMIT":
        return None, tokens

    tokens = tokens[1:]
    limit_token = tokens[0]
    assert limit_token.type == "int", f"Expected limit value, got {limit_token}"
    limit_value = limit_token.value
    limit_int = int(limit_value)
    tokens = tokens[1:]

    if tokens and tokens[0].type == "keyword" and tokens[0].value.upper() == "OFFSET":
        tokens = tokens[1:]
        offset_token = tokens[0]
        assert offset_token.type == "int", f"Expected offset value, got {offset_token}"
        offset_value = offset_token.value
        offset_int = int(offset_value)
        tokens = tokens[1:]
        return Limit(limit=limit_int, offset=offset_int), tokens
    else:
        return Limit(limit=limit_int), tokens


def parse_select(tokens: List[Token]) -> Tuple[Select, List[Token]]:
    if not tokens or tokens[0].type != "keyword" or tokens[0].value.upper() != "SELECT":
        raise ValueError("Expected SELECT statement")
    tokens = tokens[1:]
    field_parts, tokens = parse_fields(tokens)
    from_part, tokens = parse_from(tokens)
    where_part, tokens = parse_where(tokens)
    order_part, tokens = parse_order(tokens)
    limit_part, tokens = parse_limit(tokens)

    return Select(
        field_parts=field_parts,
        from_part=from_part,
        where_part=where_part,
        order_part=order_part,
        limit_part=limit_part,
    ), tokens


def parse(tokens: List[Token]) -> Ast:
    select_part, tokens = parse_select(tokens)
    if tokens:
        raise ValueError(
            "Unexpected tokens after SELECT statement. Remaining tokens: " + str(tokens)
        )

    if select_part is not None:
        return select_part
    else:
        raise ValueError("Failed to parse SELECT statement")



from unittest import TestCase
from .parser import parse, parse_expression, parse_from, parse_limit
from .lexer import scan
from .ast import (
    Select,
    From,
    SelectWildcard,
    PlusExpression,
    SelectField,
    IntExpression,
    NameExpression,
    EqualExpression,
    StringExpression,
    Where,
    OrderBy,
    Limit,
    OrderField,
)


class ParserTest(TestCase):
    def test_select_literal(self):
        tokens = scan("SELECT foo")
        ast = parse(tokens)
        self.assertEqual(
            ast,
            Select(
                field_parts=[SelectField(expression=NameExpression(name="foo"))],
                from_part=None,
            ),
        )

    def test_select_wildcard(self):
        tokens = scan("SELECT * FROM users")
        ast = parse(tokens)
        self.assertEqual(
            ast,
            Select(
                field_parts=[SelectWildcard()],
                from_part=From(
                    table="users",
                ),
            ),
        )

    def test_select_with_field(self):
        tokens = scan("SELECT name FROM users")
        ast = parse(tokens)
        self.assertEqual(
            ast,
            Select(
                field_parts=[SelectField(expression=NameExpression(name="name"))],
                from_part=From(
                    table="users",
                ),
            ),
        )

    def test_select_with_alias(self):
        tokens = scan("select foo from bar as baz")
        ast = parse(tokens)
        self.assertEqual(
            ast,
            Select(
                field_parts=[SelectField(expression=NameExpression(name="foo"))],
                from_part=From(
                    table="bar",
                    alias="baz",
                ),
            ),
        )

    def test_select_where(self):
        self.maxDiff = None
        tokens = scan("SELECT name FROM users WHERE name = 'John'")
        ast = parse(tokens)
        self.assertEqual(
            ast,
            Select(
                field_parts=[SelectField(expression=NameExpression(name="name"))],
                from_part=From(
                    table="users",
                ),
                where_part=Where(
                    expression=EqualExpression(
                        left=NameExpression(name="name"),
                        right=StringExpression(value="John"),
                    )
                ),
            ),
        )

    def test_select_order(self):
        tokens = scan("SELECT foo FROM users ORDER BY bar DESC")
        ast = parse(tokens)
        self.assertEqual(
            ast,
            Select(
                field_parts=[SelectField(expression=NameExpression(name="foo"))],
                from_part=From(
                    table="users",
                ),
                order_part=OrderBy(
                    fields=[
                        OrderField(
                            expression=NameExpression(name="bar"), direction="DESC"
                        )
                    ]
                ),
            ),
        )


class ExpressionTest(TestCase):
    def test_plus_expression(self):
        exp = scan("1+1")
        ast, tokens = parse_expression(exp)
        self.assertEqual(
            ast,
            PlusExpression(
                left=IntExpression(value=1),
                right=IntExpression(value=1),
            ),
        )
        self.assertEqual(tokens, [])

    def test_equal_expression(self):
        exp = scan("name = 'John'")
        ast, tokens = parse_expression(exp)
        self.assertEqual(
            ast,
            EqualExpression(
                left=NameExpression(name="name"),
                right=StringExpression(value="John"),
            ),
        )
        self.assertEqual(tokens, [])


class FromTest(TestCase):
    def test_simple(self):
        tokens = scan("FROM users")
        ast, tokens = parse_from(tokens)
        self.assertEqual(
            ast,
            From(
                table="users",
            ),
        )
        self.assertEqual(tokens, [])

    def test_with_alias(self):
        tokens = scan("FROM users AS u")
        ast, tokens = parse_from(tokens)
        self.assertEqual(
            ast,
            From(
                table="users",
                alias="u",
            ),
        )
        self.assertEqual(tokens, [])


class LimitTest(TestCase):
    def test_limit(self):
        tokens = scan("LIMIT 10")
        ast, tokens = parse_limit(tokens)
        self.assertEqual(
            ast,
            Limit(
                limit=10,
            ),
        )
        self.assertEqual(tokens, [])

    def test_limit_with_offset(self):
        tokens = scan("LIMIT 10 OFFSET 5")
        ast, tokens = parse_limit(tokens)
        self.assertEqual(
            ast,
            Limit(
                limit=10,
                offset=5,
            ),
        )
        self.assertEqual(tokens, [])



from dataclasses import dataclass
from typing import Literal


operators = [
    "<>",
    ">=",
    "<=",
    "=",
    ">",
    "<",
    "!=",
    "+",
    "-",
    "*",
    "/",
]

keywords = [
    "SELECT",
    "FROM",
    "WHERE",
    "AND",
    "AS",
    "OR",
    "NOT",
    "IN",
    "LIKE",
    "BETWEEN",
    "IS",
    "NULL",
    "EXISTS",
    "OFFSET",
    "DISTINCT",
    "ORDER BY",
    "GROUP BY",
    "HAVING",
    "ASC",
    "DESC",
    "INNER JOIN",
    "LEFT JOIN",
    "RIGHT JOIN",
    "FULL JOIN",
    "CROSS JOIN",
    "NATURAL JOIN",
    "JOIN",
    "LIMIT",
]


@dataclass
class Token:
    type: Literal[
        "name", "operator", "str", "int", "float", "keyword", "wildcard", "comma"
    ]
    value: str



from .ast import SelectField, NameExpression


def field_name(field: SelectField) -> str:
    """
    Get the field name from a SelectField object.
    """
    if field.alias:
        return field.alias
    elif isinstance(field.expression, NameExpression):
        return field.expression.name
    else:
        return "?column?"



from unittest import TestCase
from .ast import (
    Where,
    GreaterThanExpression,
    NameExpression,
    IntExpression,
    OrderBy,
    OrderField,
    GroupBy,
    StringExpression,
    PlusExpression,
    MinusExpression,
    MultiplyExpression,
    DivideExpression,
    FloatExpression,
    Limit,
)

from .apply import (
    apply_expression,
    apply_where,
    apply_order_by,
    apply_group_by,
    apply_limit,
)


class TestApplyExpression(TestCase):
    def test_addition(self):
        expression = PlusExpression(
            left=NameExpression(name="a"),
            right=NameExpression(name="b"),
        )
        ctx = {"a": 1, "b": 2}
        result = apply_expression(expression, ctx)
        self.assertEqual(result, 3)

    def test_subtraction(self):
        expression = MinusExpression(
            left=NameExpression(name="a"),
            right=NameExpression(name="b"),
        )
        ctx = {"a": 5, "b": 2}
        result = apply_expression(expression, ctx)
        self.assertEqual(result, 3)

    def test_multiplication(self):
        expression = MultiplyExpression(
            left=NameExpression(name="a"),
            right=NameExpression(name="b"),
        )
        ctx = {"a": 3, "b": 4}
        result = apply_expression(expression, ctx)
        self.assertEqual(result, 12)

    def test_division(self):
        expression = DivideExpression(
            left=NameExpression(name="a"),
            right=NameExpression(name="b"),
        )
        ctx = {"a": 8, "b": 2}
        result = apply_expression(expression, ctx)
        self.assertEqual(result, 4)

    def test_string(self):
        expression = StringExpression(value="foo")
        ctx = {}
        result = apply_expression(expression, ctx)
        self.assertEqual(result, "foo")

    def test_int(self):
        expression = IntExpression(value=42)
        ctx = {}
        result = apply_expression(expression, ctx)
        self.assertEqual(result, 42)

    def test_float(self):
        expression = FloatExpression(value=3.14)
        ctx = {}
        result = apply_expression(expression, ctx)
        self.assertEqual(result, 3.14)

    def test_boolean(self):
        expression = NameExpression(name="TrUe")
        ctx = {}
        result = apply_expression(expression, ctx)
        self.assertEqual(result, True)


class TestApplyWhere(TestCase):
    def test_where_no_context(self):
        where = Where(
            expression=GreaterThanExpression(
                left=NameExpression(name="age"),
                right=IntExpression(value=18),
            )
        )

        data = [
            {"name": "Alice", "age": 20},
            {"name": "Bob", "age": 17},
            {"name": "Charlie", "age": 19},
            {"name": "David", "age": 16},
            {"name": "Eve", "age": 22},
        ]

        result = apply_where(where=where, data=data, ctx={})

        self.assertEqual(
            result,
            [
                {"name": "Alice", "age": 20},
                {"name": "Charlie", "age": 19},
                {"name": "Eve", "age": 22},
            ],
        )

    def test_with_context(self):
        where = Where(
            expression=GreaterThanExpression(
                left=NameExpression(name="age"),
                right=NameExpression(name="minimum_age"),
            )
        )

        data = [
            {"name": "Alice", "age": 20},
            {"name": "Bob", "age": 17},
            {"name": "Charlie", "age": 19},
            {"name": "David", "age": 16},
            {"name": "Eve", "age": 22},
        ]

        result = apply_where(
            where=where,
            data=data,
            ctx={
                "minimum_age": 18,
            },
        )

        self.assertEqual(
            result,
            [
                {"name": "Alice", "age": 20},
                {"name": "Charlie", "age": 19},
                {"name": "Eve", "age": 22},
            ],
        )


class TestApplyOrderBy(TestCase):
    def test_order_by(self):
        order_by = OrderBy(
            fields=[
                OrderField(
                    expression=NameExpression(name="age"),
                    direction="ASC",
                )
            ]
        )

        data = [
            {"name": "Alice", "age": 20},
            {"name": "Bob", "age": 17},
            {"name": "Charlie", "age": 19},
            {"name": "David", "age": 16},
            {"name": "Eve", "age": 22},
        ]

        result = apply_order_by(order_by=order_by, data=data, ctx={})

        self.assertEqual(
            result,
            [
                {"name": "David", "age": 16},
                {"name": "Bob", "age": 17},
                {"name": "Charlie", "age": 19},
                {"name": "Alice", "age": 20},
                {"name": "Eve", "age": 22},
            ],
        )


class TestApplyGroupBy(TestCase):
    def test_group_by(self):
        group_by = GroupBy(
            fields=[
                NameExpression(name="team"),
            ],
        )

        data = [
            {"name": "Alice", "team": "foo"},
            {"name": "Bob", "team": "bar"},
            {"name": "Charlie", "team": "foo"},
            {"name": "David", "team": "bar"},
            {"name": "Eve", "team": "foo"},
        ]

        result = apply_group_by(group_by=group_by, data=data, ctx={})

        self.assertEqual(
            result,
            {
                ("foo",): [
                    {"name": "Alice", "team": "foo"},
                    {"name": "Charlie", "team": "foo"},
                    {"name": "Eve", "team": "foo"},
                ],
                ("bar",): [
                    {"name": "Bob", "team": "bar"},
                    {"name": "David", "team": "bar"},
                ],
            },
        )


class TestApplyLimit(TestCase):
    def test_limit(self):
        data = [
            {"name": "Alice", "age": 20},
            {"name": "Bob", "age": 17},
            {"name": "Charlie", "age": 19},
            {"name": "David", "age": 16},
            {"name": "Eve", "age": 22},
        ]

        limit = Limit(limit=3)
        result = apply_limit(data=data, limit=limit, ctx={})

        self.assertEqual(
            result,
            [
                {"name": "Alice", "age": 20},
                {"name": "Bob", "age": 17},
                {"name": "Charlie", "age": 19},
            ],
        )



from typing import List, Optional, Literal, Union
from dataclasses import dataclass
from abc import ABC


@dataclass
class Ast(ABC): ...


@dataclass
class Command(Ast): ...


@dataclass
class Expression(Ast): ...


@dataclass
class NameExpression(Expression):
    name: str


@dataclass
class StringExpression(Expression):
    value: str


@dataclass
class IntExpression(Expression):
    value: int


@dataclass
class FloatExpression(Expression):
    value: float


@dataclass
class PlusExpression(Expression):
    left: Expression
    right: Expression


@dataclass
class MinusExpression(Expression):
    left: Expression
    right: Expression


@dataclass
class MultiplyExpression(Expression):
    left: Expression
    right: Expression


@dataclass
class DivideExpression(Expression):
    left: Expression
    right: Expression


@dataclass
class LessThanExpression(Expression):
    left: Expression
    right: Expression


@dataclass
class LessThanOrEqualExpression(Expression):
    left: Expression
    right: Expression


@dataclass
class GreaterThanExpression(Expression):
    left: Expression
    right: Expression


@dataclass
class GreaterThanOrEqualExpression(Expression):
    left: Expression
    right: Expression


@dataclass
class EqualExpression(Expression):
    left: Expression
    right: Expression


@dataclass
class NotEqualExpression(Expression):
    left: Expression
    right: Expression


@dataclass
class Join(Ast):
    table: str
    table_alias: str
    join_type: Literal["INNER", "LEFT", "RIGHT"]
    on: Expression


@dataclass
class From(Ast):
    table: str
    alias: Optional[str] = None
    join: Optional[List[Join]] = None


@dataclass
class Where(Ast):
    expression: Expression


@dataclass
class SelectField(Ast):
    expression: Expression
    alias: Optional[str] = None


@dataclass
class OrderField(Ast):
    expression: Expression
    direction: Literal["ASC", "DESC"]


@dataclass
class OrderBy(Ast):
    fields: List[OrderField]


@dataclass
class SelectWildcard(Ast):
    pass


@dataclass
class GroupBy(Ast):
    fields: List[Expression]


@dataclass
class Limit(Ast):
    limit: int
    offset: int = 0


@dataclass
class Select(Command):
    field_parts: List[Union[SelectField, SelectWildcard]]
    from_part: Optional[From] = None
    where_part: Optional[Where] = None
    order_part: Optional[OrderBy] = None
    group_part: Optional[GroupBy] = None
    limit_part: Optional[Limit] = None



from typing import List, Tuple
from .tokens import Token, operators, keywords


def start_with_operator(code: str):
    return any(code.startswith(op) for op in operators)


def extract_operator(code: str):
    if start_with_operator(code):
        op = next(op for op in operators if code.startswith(op))
        code = code[len(op) :]
        return Token("operator", op), code
    else:
        raise Exception(f"Not a valid operator, code: {code}")


def start_with_space(code: str):
    return code[0].isspace()


def extract_space(code: str) -> str:
    if not start_with_space(code):
        raise Exception(f"Not white space, code: {code}")
    result = ""
    for idx, char in enumerate(code):
        if char.isspace():
            result = result + char
        else:
            return code[idx:]
    return ""


def start_with_name(code: str):
    return code[0].isalnum()


def extract_name(code: str):
    if not start_with_name(code):
        raise Exception(f"Not a valid name, code: {code}")
    result = ""
    for idx, char in enumerate(code):
        if char.isalnum() or any(
            k.startswith((result + char).upper()) for k in keywords
        ):
            result = result + char
        elif result.upper() in keywords:
            return Token("keyword", result), code[idx:]
        else:
            return Token("name", result), code[idx:]
    if code.upper() in keywords:
        return Token("keyword", code), ""
    else:
        return Token("name", code), ""


def start_with_quoted_name(code: str):
    return code[0] == '"'


def extract_quoted_name(code: str):
    if not start_with_quoted_name(code):
        raise Exception(f"Not a valid str, code: {code}")

    enumeration = enumerate(code)
    for idx, char in enumeration:
        next_idx = idx + 1
        next_char = code[next_idx] if next_idx < len(code) else None
        if char == '"' and '"' == next_char and idx > 0:
            next(enumeration)
        elif char == '"' and next_char != '"' and idx > 0:
            value = code[1:idx].replace('""', '"')
            return Token("name", value), code[next_idx:]


def start_with_number(code: str):
    if code[0].isnumeric():
        return True


def extract_number(code: str) -> Tuple[Token, str]:
    if not start_with_number(code):
        raise Exception(f"Not a valid number, code: {code}")
    result = ""
    result_type = "int"
    for idx, char in enumerate(code):
        if char.isalnum():
            result = result + char
        elif char == "." and result_type == "int":
            result_type = "float"
        else:
            return Token(result_type, result), code[idx:]
    return Token(result_type, code), ""


def start_with_str(code: str):
    return code[0] == "'"


def extract_str(code: str):
    if not start_with_str(code):
        raise Exception(f"Not a valid str, code: {code}")

    enumeration = enumerate(code)
    for idx, char in enumeration:
        next_idx = idx + 1
        next_char = code[next_idx] if next_idx < len(code) else None
        if char == "'" and "'" == next_char and idx > 0:
            next(enumeration)
        elif char == "'" and next_char != "'" and idx > 0:
            value = code[1:idx].replace("''", "'")
            return Token("str", value), code[next_idx:]


def start_with_wildcard(code: str):
    return code[0] == "*"


def extract_wildcard(code: str):
    if not start_with_wildcard(code):
        raise Exception(f"Not a valid wildcard, code: {code}")
    return Token("wildcard", "*"), code[1:]


def scan(code: str) -> List[Token]:
    result = []
    while len(code) > 0:
        if start_with_wildcard(code):
            token, code = extract_wildcard(code)
            result.append(token)
        elif start_with_number(code):
            token, code = extract_number(code)
            result.append(token)
        elif start_with_name(code):
            token, code = extract_name(code)
            result.append(token)
        elif start_with_quoted_name(code):
            token, code = extract_quoted_name(code)
            result.append(token)
        elif start_with_operator(code):
            token, code = extract_operator(code)
            result.append(token)
        elif start_with_str(code):
            token, code = extract_str(code)
            result.append(token)
        elif start_with_space(code):
            code = extract_space(code)
        else:
            raise Exception(f"Invalid token, code: {code}")
    return result



from .lexer import scan
from .parser import parse
from .apply import apply_command
from .tables import ITablesSnapshot


def eval_sql(code: str, tables: ITablesSnapshot, ctx: dict):
    tokens = scan(code)
    ast = parse(tokens)
    result = apply_command(command=ast, tables=tables, ctx=ctx)
    return result



from unittest import TestCase
from .eval import eval_sql
from .tables import InMemoryTables


class TestEvalSQL(TestCase):
    def test_eval_sql(self):
        code = "select 1+1"
        tables = InMemoryTables(
            tables=[],
        )
        ctx = {}
        result = eval_sql(code=code, tables=tables, ctx=ctx)
        self.assertEqual(result, [{None: 2}])

    def test_eval_select_alias(self):
        code = "select 1+1 as a"
        tables = InMemoryTables(
            tables=[],
        )
        ctx = {}
        result = eval_sql(code=code, tables=tables, ctx=ctx)
        self.assertEqual(result, [{"a": 2}])



from unittest import TestCase
from .lexer import scan, Token


class TestTokens(TestCase):
    def test_simple(self):
        code = "select a from b where c = d"
        self.assertEqual(
            scan(code),
            [
                Token("keyword", "select"),
                Token("name", "a"),
                Token("keyword", "from"),
                Token("name", "b"),
                Token("keyword", "where"),
                Token("name", "c"),
                Token("operator", "="),
                Token("name", "d"),
            ],
        )

    def test_joined_expression(self):
        code = "select 1+1"
        self.assertEqual(
            scan(code),
            [
                Token("keyword", "select"),
                Token("int", "1"),
                Token("operator", "+"),
                Token("int", "1"),
            ],
        )

    def test_joined_comparison(self):
        code = "select a=b"
        self.assertEqual(
            scan(code),
            [
                Token("keyword", "select"),
                Token("name", "a"),
                Token("operator", "="),
                Token("name", "b"),
            ],
        )

    def test_str(self):
        code = "select 'foo'='bar'"
        self.assertEqual(
            scan(code),
            [
                Token("keyword", "select"),
                Token("str", "foo"),
                Token("operator", "="),
                Token("str", "bar"),
            ],
        )

    def test_escaped_str(self):
        code = "select 'foo''s name' = 'bar''s name'"
        self.assertEqual(
            scan(code),
            [
                Token("keyword", "select"),
                Token("str", "foo's name"),
                Token("operator", "="),
                Token("str", "bar's name"),
            ],
        )

    def test_tricky_escape(self):
        code = "select 'foo''' = '''bar'"
        self.assertEqual(
            scan(code),
            [
                Token("keyword", "select"),
                Token("str", "foo'"),
                Token("operator", "="),
                Token("str", "'bar"),
            ],
        )

    def test_quoted_name(self):
        code = 'select foo from "my table"'
        self.assertEqual(
            scan(code),
            [
                Token("keyword", "select"),
                Token("name", "foo"),
                Token("keyword", "from"),
                Token("name", "my table"),
            ],
        )

    def test_escaped_quoted_name(self):
        code = 'select foo from "my ""table"""'
        self.assertEqual(
            scan(code),
            [
                Token("keyword", "select"),
                Token("name", "foo"),
                Token("keyword", "from"),
                Token("name", 'my "table"'),
            ],
        )

    def test_wildcard(self):
        code = "select * from users"
        self.assertEqual(
            scan(code),
            [
                Token("keyword", "select"),
                Token("wildcard", "*"),
                Token("keyword", "from"),
                Token("name", "users"),
            ],
        )

    def test_order_by(self):
        code = "select foo from users order by bar"
        self.assertEqual(
            scan(code),
            [
                Token("keyword", "select"),
                Token("name", "foo"),
                Token("keyword", "from"),
                Token("name", "users"),
                Token("keyword", "order by"),
                Token("name", "bar"),
            ],
        )

    def test_expression(self):
        code = "1+1"
        self.assertEqual(
            scan(code), [Token("int", "1"), Token("operator", "+"), Token("int", "1")]
        )
